"""Base Large Language Model (LLM) class."""

import logging
import re
from typing import Any, TYPE_CHECKING

from loguru import logger

if TYPE_CHECKING:
    from semantix.inference import ExtractOutputPromptInfo, OutputFixPromptInfo


httpx_logger = logging.getLogger("httpx")
httpx_logger.setLevel(logging.WARNING)

NORMAL = """
Generate and return the output result(s) only, adhering to the provided Type in the following format

[Output] <result>
"""  # noqa E501

REASON = """
Reason and return the output result(s) only, adhering to the provided Type in the following format

[Reasoning] <Reason>
[Output] <Result>
"""

CHAIN_OF_THOUGHT = """
Generate and return the output result(s) only, adhering to the provided Type in the following format. Perform the operation in a chain of thoughts.(Think Step by Step)

[Chain of Thoughts] <Chain of Thoughts>
[Output] <Result>
"""  # noqa E501

REACT = """
You are given with a list of tools you can use to do different things. To achieve the given [Action], incrementally think and provide tool_usage necessary to achieve what is thought. Provide your answer adhering in the following format. tool_usage is a function call with the necessary arguments. Only provide one [THOUGHT] and [TOOL USAGE] at a time.

[Thought] <Thought>
[Tool Usage] <tool_usage>
"""  # noqa E501


class BaseLLM:
    """Base Large Language Model (LLM) class."""

    MESSAGE_DESCRIPTIONS = {
        "informations": "[Information]",
        "input_informations": "[Inputs Information]",
        "context": "[Context]",
        "type_explanations": "[Type Explanations]",
        "return_hint": "[Output Information]",
        "action": "[Action]",
        "tools": "[Tools]",
    }
    SYSTEM_PROMPT = "This is an operation you must perform and return the output values. Neither, the methodology, extra sentences nor the code are not needed. Input/Type formatting: Explanation of the Input (variable_name) (type) = value"  # noqa E501
    METHOD_PROMPTS = {
        "Normal": NORMAL,
        "Reason": REASON,
        "Chain-of-Thoughts": CHAIN_OF_THOUGHT,
        "ReAct": REACT,
    }
    EXTRACT_OUTPUT_INSTRUCTION = "Above output is not in the desired Output Format/Type. Please provide the output in the desired type. Do not repeat the previously provided output. Important: Do not provide the code or the methodology. Only provide the output in the desired format."  # noqa E501
    OUTPUT_FIX_INSTRUCTION = "Above output is not in the desired Output Format/Type. Please provide the output in the desired type. Do not repeat the previously provided output. Important: Do not provide the code or the methodology. Only provide the output in the desired format."  # noqa E501

    def __init__(self, verbose: bool = False, max_retries: int = 3) -> None:
        """Initialize the Large Language Model (LLM) client."""
        self.verbose = verbose
        self.max_retries = max_retries

    def get_message_desc(self, key: str) -> str:
        """Get the message description."""
        return self.MESSAGE_DESCRIPTIONS[key]

    @property
    def system_message(self) -> dict:
        """Get the system message."""
        return {"role": "system", "content": self.SYSTEM_PROMPT}

    def method_message(self, method: str) -> dict:
        """Get the method message."""
        return {"role": "system", "content": self.METHOD_PROMPTS[method]}

    def __infer__(self, messages: list, model_params: dict) -> str:
        """Infer a response from the input meaning."""
        raise NotImplementedError

    @staticmethod
    def _msgs_to_str(messages: list) -> str:
        """Convert the messages to a string."""
        return "\n\n".join([str(m["content"]) for m in messages])

    def __call__(self, messages: list, model_params: dict) -> str:
        """Infer a response from the input text."""
        if self.verbose:
            logger.info(f"Model Input\n{self._msgs_to_str(messages)}")
        return self.__infer__(messages, model_params)

    def resolve_output(
        self,
        model_output: str,
        extract_output_prompt_info: "ExtractOutputPromptInfo",
        output_fix_prompt_info: "OutputFixPromptInfo",
        _globals: dict,
        _locals: dict,
    ) -> Any:  # noqa: ANN401
        """Resolve the output string to return the reasoning and output."""
        if self.verbose:
            logger.info(f"Model Output\n{model_output}")
        output_match = re.search(r"\[Output\](.*)", model_output, re.DOTALL)
        if not output_match:
            output = self._extract_output(
                model_output,
                extract_output_prompt_info,
            )
        else:
            output = output_match.group(1).strip()
        return self.to_object(output, output_fix_prompt_info, _globals, _locals)

    def _extract_output(
        self, model_output: str, extract_output_prompt_info: "ExtractOutputPromptInfo"
    ) -> str:
        """Extract the output from the meaning out string."""
        if self.verbose:
            logger.info("Extracting output from the meaning out string.")
        output_extract_messages = extract_output_prompt_info.get_messages(
            self, model_output
        )
        output_extract_output = self.__infer__(output_extract_messages, {})
        if self.verbose:
            logger.info(f"Extracted Output: {output_extract_output}")
        return output_extract_output

    def to_object(
        self,
        output: str,
        output_fix_prompt_info: "OutputFixPromptInfo",
        _globals: dict,
        _locals: dict,
        error: str = "",
        num_retries: int = 0,
    ) -> Any:  # noqa: ANN401
        """Convert the output string to an object."""
        if output_fix_prompt_info.return_hint.type == "str":
            return output
        if num_retries >= self.max_retries:
            raise ValueError("Failed to convert output to object. Max tries reached.")
        if error:
            fixed_output = self._fix_output(output, output_fix_prompt_info, error)
            return self.to_object(
                fixed_output,
                output_fix_prompt_info,
                _globals,
                _locals,
                error="",
                num_retries=num_retries + 1,
            )
        try:
            return eval(output, _globals, _locals)
        except Exception as e:
            return self.to_object(
                output,
                output_fix_prompt_info,
                _globals,
                _locals,
                error=str(e),
                num_retries=num_retries + 1,
            )

    def _fix_output(
        self, output: str, output_fix_prompt_info: "OutputFixPromptInfo", error: str
    ) -> str:
        """Fix the output string."""
        if self.verbose:
            logger.info(f"Error: {error}, Fixing the output.")
        output_fix_messages = output_fix_prompt_info.get_messages(self, output, error)
        output_fix_output = self.__infer__(output_fix_messages, {})
        return output_fix_output



"""This module contains the classes and functions to represent the types and information needed for the library."""

import inspect
import sys
from enum import Enum
from types import FrameType
from typing import Any, Callable, Dict, Generic, List, Type, TypeVar, Union

from semantix.utils import (
    extract_non_primary_type,
    get_object_string,
    get_type,
    get_type_from_value,
)


T = TypeVar("T")


class SemanticMeta(type):
    """Metaclass for the Semantic class."""

    def __new__(
        mcs, name: str, bases: tuple, namespace: dict, **kwargs: dict  # noqa: N804
    ) -> Any:  # noqa: ANN401
        """Creates a new instance of the class."""
        cls = super().__new__(mcs, name, bases, namespace)
        if "meaning" in kwargs and hasattr(cls, "meaning"):
            cls._meaning = kwargs["meaning"]  # type: ignore
        return cls

    def __getitem__(cls, params: tuple) -> Type[T]:
        """Get the item from the class."""
        if not isinstance(params, tuple) or len(params) != 2:
            raise TypeError("Semantic requires two parameters: type and meaning")
        typ, meaning = params
        curr_frame = inspect.currentframe()
        if curr_frame:
            frame = curr_frame.f_back
        if not frame:
            raise Exception("Cannot get the current frame.")
        var_name = list(frame.f_locals.keys())[-1]
        # Set the meaning of the variable in the module's global scope
        if var_name:
            setattr(
                sys.modules[frame.f_globals["__name__"]], f"{var_name}_meaning", meaning
            )
        return type(
            f"MT_{get_type(typ)}", (cls,), {"wrapped_type": typ, "_meaning": meaning}
        )


class Semantic(Generic[T], metaclass=SemanticMeta):
    """Class to represent the semantic type."""

    wrapped_type: Type[T]
    _meaning: str = ""

    def __new__(cls, *args: list, **kwargs: dict) -> Any:  # noqa: ANN401
        """Creates a new instance of the class."""
        return cls.wrapped_type(*args, **kwargs)

    def __instancecheck__(self, instance: Any) -> bool:  # noqa: ANN401
        """Check if the instance is of the class."""
        return isinstance(instance, self.wrapped_type)

    def __subclasscheck__(self, subclass: Type) -> bool:
        """Check if the subclass is of the class."""
        return issubclass(subclass, self.wrapped_type)

    def __repr__(self) -> str:
        """Get the representation of the class."""
        return f"{self.wrapped_type.__name__} {self._meaning}"


class SemanticClass:
    """Class to represent the semantic class."""

    @classmethod
    def init(cls, *args: list, **kwargs: dict) -> Any:  # noqa: ANN401
        """Initialize the class."""
        # TODO: Implement the initialization of the class with llms
        return cls.__class__(*args, **kwargs)


class TypeExplanation:
    """Class to represent the type explanation."""

    def __init__(self, frame: FrameType, type: str) -> None:
        """Initializes the TypeExplanation class."""
        self.type = frame.f_globals[type]

    def get_type_repr(self, type_collector: list = []) -> str:
        """Get the type representation."""
        semstr = self.type.__doc__ if self.type.__doc__ else ""
        _name = self.type.__name__
        usage_example_list = []
        for param, annotation in self.type.__init__.__annotations__.items():
            if param == "return":
                continue
            if isinstance(annotation, type) and issubclass(annotation, Semantic):
                type_repr = get_type(annotation.wrapped_type)
                type_collector.extend(extract_non_primary_type(type_repr))
                usage_example_list.append(
                    f'{param}="{annotation._meaning}":{type_repr}'
                )
            else:
                type_repr = get_type(annotation)
                type_collector.extend(extract_non_primary_type(type_repr))
                usage_example_list.append(f"{param}={type_repr}")
        usage_example = ", ".join(usage_example_list)
        return f"{semstr} ({_name}) (class) eg:- {_name}({usage_example})".strip()

    def get_type_repr_enum(self) -> str:
        """Get the type representation."""
        semstr = self.type.__doc__ if self.type.__doc__ else ""
        _name = self.type.__name__
        usage_example_list = []
        for param, _ in self.type.__members__.items():
            usage_example_list.append(f"{_name}.{param}")
        usage_example = ", ".join(usage_example_list)
        return f"{semstr} ({_name}) (enum) eg:- {usage_example}".strip()

    def __str__(self) -> str:
        """Returns the string representation of the TypeExplanation class."""
        if issubclass(self.type, Enum):
            return self.get_type_repr_enum()
        return self.get_type_repr()

    def get_nested_types(self) -> list:
        """Get the nested types."""
        type_collector: List[str] = []
        if not issubclass(self.type, Enum):
            self.get_type_repr(type_collector)
        return type_collector


class Information:
    """Class to represent the information."""

    def __init__(self, semstr: str, name: str, value: Any) -> None:  # noqa: ANN401
        """Initializes the Information class."""
        self.value = value
        self.name = name
        self.semstr = semstr

    @property
    def type(self) -> str:
        """Get the type of the information."""
        return get_type_from_value(self.value)

    def get_content(self, contains_media: bool) -> Union[List[Dict], str]:
        """Returns the list of dictionaries representation of the InputInformation class."""
        input_type = self.type
        if input_type == "Image":
            img_base64, img_type = self.value.process()
            return [
                {
                    "type": "text",
                    "text": f"{self.semstr if self.semstr else ''} ({self.name}) (Image) = ".strip(),
                },
                {
                    "type": "image_url",
                    "image_url": {"url": f"data:image/{img_type};base64,{img_base64}"},
                },
            ]
        elif input_type == "Video":
            video_frames = self.value.process()
            return [
                {
                    "type": "text",
                    "text": f"{self.semstr if self.semstr else ''} ({self.name}) (Video) = ".strip(),
                },
                *(
                    {
                        "type": "image_url",
                        "image_url": {
                            "url": f"data:image/jpeg;base64,{frame}",
                            "detail": "low",
                        },
                    }
                    for frame in video_frames
                ),
            ]
        return (
            str(self)
            if not contains_media
            else [
                {
                    "type": "text",
                    "text": str(self),
                }
            ]
        )

    def __str__(self) -> str:
        """Returns the string representation of the Information class."""
        return f"{self.semstr} ({self.name}) ({self.type}) = {get_object_string(self.value)}".strip()

    def get_types(self) -> list:
        """Get the types of the information."""
        type_collector = extract_non_primary_type(self.type)
        get_object_string(self.value, type_collector)
        return type_collector


class OutputHint:
    """Class to represent the output hint."""

    def __init__(self, semstr: str, type: Type[Any]) -> None:  # noqa: ANN401
        """Initializes the OutputHint class."""
        self.semstr = semstr
        self.type = get_type(type)

    def __str__(self) -> str:
        """Returns the string representation of the OutputHint class."""
        return f"{self.semstr if self.semstr else ''} ({self.type})".strip()

    def get_types(self) -> list:
        """Get the types of the output."""
        type_collector = extract_non_primary_type(self.type)
        return type_collector


class Tool:
    """Base class for tools."""

    def __init__(self, func: Callable, semstr: str = "") -> None:
        """Initialize the tool."""
        self.func = func
        self.semstr = semstr

    @property
    def get_params(self) -> List[Dict]:
        """Get the parameters of the tool."""
        params = []
        for param, annotation in self.func.__annotations__.items():
            if isinstance(annotation, type) and issubclass(annotation, Semantic):
                params.append(
                    {
                        "name": param,
                        "type": get_type(annotation.wrapped_type),
                        "semstr": annotation._meaning,
                    }
                )
            else:
                params.append(
                    {
                        "name": param,
                        "type": get_type(annotation),
                        "semstr": "",
                    }
                )
        return params

    def __call__(self, *args, **kwargs) -> str:  # noqa
        """Forward function of the tool."""
        return self.func(*args, **kwargs)

    def get_usage_example(self) -> str:
        """Get the usage example of the tool."""
        get_param_str = lambda x: (  # noqa E731
            f'{x["name"]}="{x["semstr"]}":{x["type"]}'
            if x["semstr"]
            else f'{x["name"]}={x["type"]}'
        )
        return f"{self.func.__name__}({', '.join([get_param_str(x) for x in self.get_params if x['name'] != 'return'])})"  # noqa E501

    def get_return_annotation(self) -> str:
        """Get the return annotation of the tool."""
        return_annotation = self.func.__annotations__.get("return")
        if (
            return_annotation
            and isinstance(return_annotation, type)
            and issubclass(return_annotation, Semantic)
        ):
            return f'returns "{return_annotation._meaning}":{get_type(return_annotation.wrapped_type)}'
        return f"returns {get_type(return_annotation)}"

    def __str__(self) -> str:
        """String representation of the tool."""
        return " | ".join(
            [
                f"{self.semstr} ({self.func.__name__})",
                self.get_return_annotation(),
                f"usage eg. {self.get_usage_example()}",
            ]
        )


class ReActOutput:
    """Class to represent the ReAct output."""

    def __init__(self, thought: str, action: str, observation: str) -> None:
        """Initializes the ReActOutput class."""
        self.thought = thought
        self.action = action
        self.observation = observation

    def __repr__(self) -> str:
        """Returns the string representation of the ReActOutput class."""
        return f"ReActOutput(thought={self.thought}, action={self.action}, observation={self.observation})"



"""Semantix is a Python library that give superpowers to your code."""

from semantix.decorators import tool, with_llm
from semantix.media import Image, Video
from semantix.types import Semantic, SemanticClass

__all__ = ["Semantic", "with_llm", "tool", "Image", "Video", "SemanticClass"]



"""Inference engine for running the model and generating prompts."""

from types import FrameType
from typing import List

from semantix.llms.base import BaseLLM
from semantix.types import Information, OutputHint, Tool, TypeExplanation


class PromptInfo:
    """Class to represent the prompt information. (According to Meaning-Typed Prompting Technique)."""

    def __init__(
        self,
        action: str,
        context: str,
        informations: List[Information],
        input_informations: List[Information],
        tools: List[Tool],
        return_hint: OutputHint,
        type_explanations: List[TypeExplanation],
    ) -> None:
        """Initializes the PromptInfo class."""
        self.informations = informations
        self.context = context
        self.input_informations = input_informations
        self.return_hint = return_hint
        self.type_explanations = type_explanations
        self.action = action
        self.tools = tools

    def get_messages(self, model: BaseLLM) -> list:
        """Get the messages for the prompt."""
        messages = [model.system_message]
        if self.informations:
            messages.append(self.get_info_msg(self.informations, model, "informations"))
        if self.input_informations:
            messages.append(
                self.get_info_msg(self.input_informations, model, "input_informations")
            )
        if self.context:
            messages.append(
                {
                    "role": "user",
                    "content": f"{model.get_message_desc('context')}\n{self.context}",
                }
            )
        if self.type_explanations:
            messages.append(
                {
                    "role": "user",
                    "content": "\n".join(
                        [
                            model.get_message_desc("type_explanations"),
                            *[str(t) for t in self.type_explanations],
                        ]
                    ),
                }
            )
        messages.append(
            {
                "role": "user",
                "content": f"{model.get_message_desc('return_hint')}\n{self.return_hint}",
            }
        )
        messages.append(
            {
                "role": "user",
                "content": f"{model.get_message_desc('action')}\n{self.action}",
            }
        )
        if self.tools:
            messages.append(
                {
                    "role": "user",
                    "content": "\n".join(
                        [model.get_message_desc("tools"), *[str(t) for t in self.tools]]
                    ),
                }
            )
        return messages

    def get_info_msg(
        self, inputs: List[Information], model: BaseLLM, info_type: str
    ) -> dict:
        """Get the information message."""
        contains_media = any(i.type == "Video" or i.type == "Image" for i in inputs)
        contents = [
            (
                model.get_message_desc(info_type)
                if not contains_media
                else {
                    "type": "text",
                    "text": model.get_message_desc(info_type),
                }
            )
        ]
        for i in inputs:
            content = i.get_content(contains_media)
            if isinstance(content, list):
                contents.extend(content)
            else:
                contents.append(content)
        return {
            "role": "user",
            "content": contents if contains_media else "\n".join(contents),  # type: ignore
        }


class ExtractOutputPromptInfo:
    """Class to represent the extract output prompt information."""

    def __init__(
        self, return_hint: OutputHint, type_explanations: List[TypeExplanation]
    ) -> None:
        """Initializes the ExtractOutputPromptInfo class."""
        self.return_hint = return_hint
        self.type_explanations = type_explanations

    def get_messages(self, model: BaseLLM, output: str) -> list:
        """Get the messages for the extract output prompt."""
        messages = []
        messages.append(
            {
                "role": "user",
                "content": f"{model.get_message_desc('extract_output_output')}\n{output}",
            }
        )
        messages.append(
            {
                "role": "user",
                "content": f"{model.get_message_desc('extract_output_return_hint')}\n{self.return_hint}",
            }
        )
        if self.type_explanations:
            messages.append(
                {
                    "role": "user",
                    "content": "\n".join(
                        [
                            model.get_message_desc("extract_output_type_explanations"),
                            *[str(t) for t in self.type_explanations],
                        ]
                    ),
                }
            )
        messages.append(
            {
                "role": "user",
                "content": model.EXTRACT_OUTPUT_INSTRUCTION,
            }
        )
        return messages


class OutputFixPromptInfo:
    """Class to represent the output fix prompt information."""

    def __init__(
        self, return_hint: OutputHint, type_explanations: List[TypeExplanation]
    ) -> None:
        """Initializes the OutputFixPromptInfo class."""
        self.return_hint = return_hint
        self.type_explanations = type_explanations

    def get_messages(self, model: BaseLLM, output: str, error: str) -> list:
        """Get the messages for the output fix prompt."""
        messages = []
        messages.append(
            {
                "role": "user",
                "content": f"{model.get_message_desc('output_fix_output')}\n{output}",
            }
        )
        messages.append(
            {
                "role": "user",
                "content": f"{model.get_message_desc('output_fix_return_hint')}\n{self.return_hint}",
            }
        )
        if self.type_explanations:
            messages.append(
                {
                    "role": "user",
                    "content": "\n".join(
                        [
                            model.get_message_desc("output_fix_type_explanations"),
                            *[str(t) for t in self.type_explanations],
                        ]
                    ),
                }
            )
        messages.append(
            {
                "role": "user",
                "content": f"{model.get_message_desc('output_fix_error')}\n{error}",
            }
        )
        messages.append(
            {
                "role": "user",
                "content": model.OUTPUT_FIX_INSTRUCTION,
            }
        )
        return messages


class InferenceEngine:
    """Class to represent the inference engine."""

    def __init__(
        self,
        model: BaseLLM,
        method: str,
        prompt_info: PromptInfo,
        extract_output_prompt_info: ExtractOutputPromptInfo,
        output_fix_prompt_info: OutputFixPromptInfo,
        model_params: dict,
    ) -> None:
        """Initializes the InferenceEngine class."""
        self.model = model
        self.method = method
        self.prompt_info = prompt_info
        self.extract_output_prompt_info = extract_output_prompt_info
        self.output_fix_prompt_info = output_fix_prompt_info
        self.model_params = model_params

    def run(self, frame: FrameType) -> str:  # noqa: ANN401
        """Run the inference engine."""
        messages = self.prompt_info.get_messages(self.model)
        messages.append(self.model.method_message(self.method))
        _locals = frame.f_locals
        _globals = frame.f_globals
        model_output = self.model(messages, self.model_params)
        return self.model.resolve_output(
            model_output,
            self.extract_output_prompt_info,
            self.output_fix_prompt_info,
            _globals,
            _locals,
        )



"""Decorators for defining semantic types and tools."""

import inspect
from typing import Any, Callable, List, Union

from semantix.inference import (
    ExtractOutputPromptInfo,
    InferenceEngine,
    OutputFixPromptInfo,
    PromptInfo,
)
from semantix.llms.base import BaseLLM
from semantix.types import Information, OutputHint, Semantic, Tool, TypeExplanation
from semantix.utils import get_semstr


def with_llm(
    meaning: str,
    model: BaseLLM,
    info: list = [],
    method: str = "Normal",
    tools: List[Union[Callable, Tool]] = [],
    model_params: dict = {},
) -> Callable:
    """Converts a function into a semantic function with LLM capabilities."""
    curr_frame = inspect.currentframe()
    if curr_frame:
        frame = curr_frame.f_back
    else:
        raise Exception(
            "Cannot get the current frame."
        )  # Don't know whether this will happen
    if not frame:
        raise Exception(
            "Cannot get the previous frame."
        )  # Don't know whether this will happen

    def decorator(func: Callable) -> Callable:
        def wrapper(**kwargs: dict) -> Any:  # noqa
            informations = [Information(obj, *get_semstr(frame, obj)) for obj in info]
            _tools = [tool if isinstance(tool, Tool) else Tool(tool) for tool in tools]
            input_informations = []
            return_hint: OutputHint
            for param, annotation in func.__annotations__.items():
                if isinstance(annotation, type) and issubclass(annotation, Semantic):
                    if param == "return":
                        return_hint = OutputHint(
                            annotation._meaning, annotation.wrapped_type
                        )
                        continue
                    input_informations.append(
                        Information(annotation._meaning, param, kwargs[param])
                    )
                else:
                    if param == "return":
                        return_hint = OutputHint("", annotation)
                        continue
                    input_informations.append(Information("", param, kwargs[param]))
            assert (
                return_hint
            ), "Return type is not defined. Please define the return type."
            action = f"{meaning} ({func.__name__})"
            context = func.__doc__ if func.__doc__ else ""

            types = set()
            for i in [
                *informations,
                *input_informations,
                return_hint if return_hint else [],
            ]:
                types.update(i.get_types())  # type: ignore
            type_explanations = [TypeExplanation(frame, t) for t in types]
            for t in type_explanations:
                types.update(t.get_nested_types())
            type_explanations = [TypeExplanation(frame, t) for t in types]

            inference_engine = InferenceEngine(
                model=model,
                method=method,
                prompt_info=PromptInfo(
                    action=action,
                    context=context,
                    informations=informations,
                    input_informations=input_informations,
                    tools=_tools,
                    return_hint=return_hint,
                    type_explanations=type_explanations,
                ),
                extract_output_prompt_info=ExtractOutputPromptInfo(
                    return_hint=return_hint, type_explanations=type_explanations
                ),
                output_fix_prompt_info=OutputFixPromptInfo(
                    return_hint=return_hint,
                    type_explanations=type_explanations,
                ),
                model_params=model_params,
            )
            return inference_engine.run(frame)

        return wrapper

    return decorator


def tool(meaning: str) -> Callable:
    """Converts a function into a tool."""

    def decorator(func: Callable) -> Tool:
        return Tool(func, meaning)

    return decorator
